from collections import deque

import numpy as np

from .maze_solver import MazeSolver


class ShortestPath(MazeSolver):
    """A maze-solving algorithm that finds the shortest path using breadth-first search."""

    def __init__(self):
        pass

    def solve(
        self, grid: np.ndarray, start: tuple[int, int], goal: set[tuple[int, int]]
    ) -> list[tuple[int, int]]:
        """Solve the maze.

        Parameters
        ----------
        grid : numpy.ndarray
            A two-dimensional array of cells representing a rectangular maze.
        start : tuple[int, int]
            The location of the start cell.
        goal : set[tuple[int, int]]
            The location(s) of the goal cell(s).

        Returns
        -------
        list[tuple[int, int]]
            An ordered list of cell locations representing the solution path.
        """
        queue = deque([start])
        visited = set()
        path = {}
        index_delta = ((-1, 0), (1, 0), (0, 1), (0, -1))

        # Loop until the queue is empty.
        while queue:
            # Get the next cell.
            current = queue.popleft()

            # Save the solution path and break the loop if the current cell is the goal.
            if current in goal:
                solution_path = []

                # Loop until the current cell is the start cell.
                while current is not start:
                    # Append current cell to the solution path.
                    solution_path.append(current)

                    # Get the next cell.
                    current = path[current]
                else:
                    solution_path.append(start)

                # Reverse the order of the list to get a proper solution path.
                solution_path.reverse()

                return solution_path

            # Add the current cell to the visited list.
            visited.add(current)

            for direction, wall in enumerate(grid[current[0]][current[1]]):
                if not wall:
                    neighbor = (
                        current[0] + index_delta[direction][0],
                        current[1] + index_delta[direction][1],
                    )

                    # If the neighbor is not in the visited list.
                    if neighbor not in visited:
                        queue.append(neighbor)  # Add the neighbor to the queue.
                        path[neighbor] = current  # Store the path.



from matplotlib import collections, colors, cm, patches
import matplotlib.pyplot as plt
import numpy as np


class Utilities:
    """A class to perform maze-related utility functions.

    Attributes
    ----------
    figure : :class:`~matplotlib.figure.Figure`, optional
        A :class:`~matplotlib.figure.Figure` instance.
    axes : :class:`~matplotlib.axes.Axes`, optional
        A :class:`~matplotlib.axes.Axes` instance.
    """

    def __init__(self):
        self.figure: plt.Figure | None = None
        self.axes: plt.Axes | None = None

    def _plot_walls(self, grid):
        """Plot the walls of the maze with Matplotlib."""
        for row in range(len(grid)):
            for column, walls in enumerate(grid[row]):
                if walls[0]:
                    self.axes.plot(
                        [column, column + 1],
                        [-row, -row],
                        color="k",
                    )
                if walls[2]:
                    self.axes.plot(
                        [column + 1, column + 1],
                        [-row, -row - 1],
                        color="k",
                    )
                if walls[3]:
                    self.axes.plot(
                        [column, column],
                        [-row, -row - 1],
                        color="k",
                    )
                if walls[1]:
                    self.axes.plot(
                        [column, column + 1],
                        [-row - 1, -row - 1],
                        color="k",
                    )

    def _initiate_plot(self, size=(8, 8)):
        """Initiate a plot from Matplotlib."""
        self.figure = plt.figure(figsize=size)
        self.axes = plt.axes()
        self.axes.set_aspect("equal")
        self.axes.set_axis_off()

    def show_grid(self, grid: np.ndarray):
        """Display a plot of a rectangular, two-dimensional maze.

        Visualization is done with `Matplotlib <https://matplotlib.org/>`_.

        Parameters
        ----------
        grid : numpy.ndarray
            A two-dimensional array of cells representing a rectangular maze.
        """
        self._initiate_plot()
        self._plot_walls(grid)
        plt.show()

    def save_grid(
        self, grid: np.ndarray, file_name: str, cell_size: int = 15, line_width: int = 2
    ):
        """Save a maze as an SVG file.

        Parameters
        ----------
        grid : numpy.ndarray
            A two-dimensional array of cells representing a rectangular maze.
        file_name : str
            A path wherein the SVG file is saved.
        cell_size : int
            The size of each cell in pixels.
        line_width : int
            The width of the wall lines in pixels.
        """
        with open(file_name, "w") as file:
            file.write(
                f'<svg xmlns="http://www.w3.org/2000/svg" width="{cell_size * len(grid[0]) + line_width}" height="{cell_size * len(grid) + line_width}" fill="none" stroke="#000" stroke-width="{line_width}" stroke-linecap="square">\n'
            )
            for row in range(len(grid)):
                for column, walls in enumerate(grid[row]):
                    a = column * cell_size + line_width / 2
                    b = row * cell_size + line_width / 2
                    c = (column + 1) * cell_size + line_width / 2
                    d = (row + 1) * cell_size + line_width / 2
                    if row == 0 and walls[0]:
                        file.write(f'\t<line x1="{a}" y1="{b}" x2="{c}" y2="{b}" />\n')
                    if walls[1]:
                        file.write(f'\t<line x1="{a}" y1="{d}" x2="{c}" y2="{d}" />\n')
                    if walls[2]:
                        file.write(f'\t<line x1="{c}" y1="{b}" x2="{c}" y2="{d}" />\n')
                    if column == 0 and walls[3]:
                        file.write(f'\t<line x1="{a}" y1="{b}" x2="{a}" y2="{d}" />\n')
            file.write("</svg>")

    def show_solution(self, grid: np.ndarray, solution_path: list[tuple[int, int]]):
        """Display a plot of a rectangular, two-dimensional maze and its solution path.

        Visualization is done with `Matplotlib <https://matplotlib.org/>`_.

        Parameters
        ----------
        grid : numpy.ndarray
            A two-dimensional array of cells representing a rectangular maze.
        solution_path : list[tuple[int, int]]
            An ordered list of cell locations representing the solution path.
        """
        self._initiate_plot()
        self._plot_walls(grid)

        # Add an ordered list of rectangle patches.
        patch_list = []
        for row, column in solution_path:
            patch_list.append(patches.Rectangle((column, -row - 1), 1, 1))

        # Create a value array for the color map.
        values = [i for i in range(len(solution_path))]
        norm = colors.Normalize().autoscale(values)

        # Add the list of patches to a patch collection.
        collection = collections.PatchCollection(patch_list, norm=norm, cmap="RdYlGn")

        # Set the value array of the patches.
        collection.set_array(values)

        # Add the collection to the axes.
        self.axes.add_collection(collection)
        plt.show()

    def save_solution(
        self,
        grid: np.ndarray,
        solution_path: list[tuple[int, int]],
        file_name: str,
        cell_size: int = 15,
        line_width: int = 2,
        colormap: str = "RdYlGn",
    ):
        """Save a maze and its solution as an SVG file.

        For more colormap selection, click `here <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_.

        Parameters
        ----------
        grid : numpy.ndarray
            A two-dimensional array of cells representing a rectangular maze.
        solution_path : list[tuple[int, int]]
            An ordered list of cell locations representing the solution path.
        file_name : str
            A path wherein the SVG file is saved.
        cell_size : int
            The size of each cell in pixels.
        line_width : int
            The width of the wall lines in pixels.
        colormap : str
            A colormap included with Matplotlib.
        """

        colormap_ = cm.get_cmap(colormap)
        array = np.linspace(0, 1, len(solution_path))
        color_list = list()

        for value in array:
            rgba = colormap_(value)
            color_list.append(colors.to_hex(rgba))

        with open(file_name, "w") as file:
            file.write(
                f'<svg xmlns="http://www.w3.org/2000/svg" width="{cell_size * len(grid[0]) + line_width}" height="{cell_size * len(grid) + line_width}">\n'
            )
            for i, cell in enumerate(solution_path):
                file.write(
                    f'\t<path fill="{color_list[i]}" d="M{cell[1] * cell_size + line_width / 2} {cell[0] * cell_size + line_width / 2}h{cell_size}v{cell_size}h-{cell_size}z" />\n'
                )
            file.write(f'\t<g fill="none" stroke="#000" stroke-width="{line_width}" stroke-linecap="square">')
            for row in range(len(grid)):
                for column, walls in enumerate(grid[row]):
                    a = column * cell_size + line_width / 2
                    b = row * cell_size + line_width / 2
                    c = (column + 1) * cell_size + line_width / 2
                    d = (row + 1) * cell_size + line_width / 2
                    if row == 0 and walls[0]:
                        file.write(
                            f'\t\t<line x1="{a}" y1="{b}" x2="{c}" y2="{b}" />\n'
                        )
                    if walls[1]:
                        file.write(
                            f'\t\t<line x1="{a}" y1="{d}" x2="{c}" y2="{d}" />\n'
                        )
                    if walls[2]:
                        file.write(
                            f'\t\t<line x1="{c}" y1="{b}" x2="{c}" y2="{d}" />\n'
                        )
                    if column == 0 and walls[3]:
                        file.write(
                            f'\t\t<line x1="{a}" y1="{b}" x2="{a}" y2="{d}" />\n'
                        )
            file.write("</g>")
            file.write("</svg>")


import numpy as np


class MazeSolver:
    """A base class for maze-solving algorithms."""

    def solve(self, grid: np.ndarray, start: tuple[int, int], goal: tuple[int, int]):
        """An abstract method to solve a maze.

        Parameters
        ----------
        grid : numpy.ndarray
            A two-dimensional array of cells representing a rectangular maze.
        start : tuple[int, int]
            The location of the start cell.
        goal : set[tuple[int, int]]
            The location(s) of the goal cell(s).

        Raises
        ------
        NotImplementedError
            If the method has not been implemented in a subclass.
        """
        raise NotImplementedError(
            "The 'solve()' method must be implemented in the subclass."
        )



import random

import numpy as np

from .algorithms import MazeGenerator, MazeSolver, RecursiveBacktracking, ShortestPath


class Maze:
    """A class to represent a rectangular, two-dimensional maze.

    Attributes
    -----------
    rows : int
        The total number of rows in the maze. Defaults to ``3``.
    columns : int
        The total number of columns in the maze. Defaults to ``3``.
    grid : numpy.ndarray
        A two-dimensional array of cells representing a rectangular maze.
    grid_size : int
        The total number of cells in the maze.
    solution_path : list[tuple[int, int]]
        An ordered list of cell locations representing the solution path.
    start : tuple[int, int]
        The location of the start cell.
    goal : set[tuple[int, int]]
        The location(s) of the goal cell(s).
    path : str, optional
        A path to a maze file. Defaults to :obj:`None`.
    generator : MazeGenerator
        An instance of a :class:`.MazeGenerator` subclass used for generating mazes. Defaults to :class:`.RecursiveBacktracking`.
    solver : MazeSolver
        An instance of a :class:`.MazeSolver` subclass used for solving mazes. Defaults to :class:`.ShortestPath`.
    """

    def __init__(
        self,
        rows: int = 3,
        columns: int = 3,
        path: str | None = None,
        generator: MazeGenerator = RecursiveBacktracking(),
        solver: MazeSolver = ShortestPath(),
    ):
        self.generator = generator
        self.solver = solver

        if path is not None:
            self.load_maze(path)
        else:
            self.rows = rows
            self.columns = columns
            self.grid_size = rows * columns
            self.grid = self.generator.generate(rows, columns)
            self.start = self.get_random_cell()
            self.goal = {self.get_random_cell()}

        self.solution_path = self.solver.solve(self.grid, self.start, self.goal)

    def load_maze(self, path: str):
        """Parse a maze file.

        Parameters
        ----------
        path : str
            A path to a maze file.
        """
        with open(path, "r") as file:
            # Parse the file into a list of strings of lines.
            lines = [line.strip() for line in file if not line.isspace()]

            # Update the attributes.
            self.rows = len(lines) // 2
            self.columns = len(lines[0]) // 4
            self.grid_size = self.rows * self.columns
            self.grid = np.full((self.rows, self.columns, 4), [False] * 4)

            # Iterate over each possible cell position and update the wall details of the current cell.
            for row in range(self.rows):
                for column in range(self.columns):
                    # If the current cell is the start.
                    if lines[row * 2 + 1][column * 4 + 2] == "S":
                        self.set_start_cell(row, column)

                    # If the current cell is the goal.
                    elif lines[row * 2 + 1][column * 4 + 2] == "G":
                        if hasattr(self, "goal"):
                            self.add_goal_cell(row, column)
                        else:
                            self.set_goal_cell(row, column)

                    # If a north wall exists.
                    if lines[row * 2][column * 4 + 2] == "-":
                        self.grid[row][column][0] = True

                    # If a south wall exists.
                    if lines[row * 2 + 2][column * 4 + 2] == "-":
                        self.grid[row][column][1] = True

                    # If an east wall exists.
                    if lines[row * 2 + 1][column * 4 + 4] == "|":
                        self.grid[row][column][2] = True

                    # If a west wall exists.
                    if lines[row * 2 + 1][column * 4] == "|":
                        self.grid[row][column][3] = True

    def generate(self, rows: int, columns: int):
        """Generate a new maze and overwrite to :attr:`grid`.

        Parameters
        ----------
        rows : int
            The total number of rows of the maze.
        columns : int
            The total number of columns of the maze.
        """
        self.rows = rows
        self.columns = columns
        self.grid_size = rows * columns
        self.grid = self.generator.generate(rows, columns)

    def solve(self):
        """Solve the maze with a specific configuration."""
        self.solution_path = self.solver.solve(self.grid, self.start, self.goal)

    def set_start_cell(self, row: int, column: int):
        """Set a cell at a location as a start cell.

        Parameters
        ----------
        row : int
            The row of a cell.
        column : int
            The column of a cell.
        """
        self.start = (row, column)

    def set_goal_cell(self, row: int, column: int):
        """Set a cell at a location as a goal cell.

        Parameters
        ----------
        row : int
            The row of a cell.
        column : int
            The column of a cell.
        """
        self.goal = {(row, column)}

    def add_goal_cell(self, row: int, column: int):
        """Add a cell at a location as a goal cell.

        Parameters
        ----------
        row : int
            The row of a cell.
        column : int
            The column of a cell.
        """
        self.goal.add((row, column))

    def get_random_cell(self) -> tuple[int, int]:
        """Get a random cell location.

        Returns
        -------
        tuple[int, int]
            The location of a random cell.
        """
        return (random.randrange(self.rows), random.randrange(self.columns))

    def remove_wall(self, cell: tuple[int, int], neighbor: tuple[int, int]):
        """Remove the wall between a cell and its neighbor.

        Parameters
        ----------
        cell : tuple[int, int]
            The location of the cell.
        neighbor : tuple[int, int]
            The location of the cell's neighbor.
        """
        if cell[1] == neighbor[1]:  # If both cells share the same column.
            # +---+
            # | C |
            # +---+
            # | N |
            # +---+
            if cell[0] < neighbor[0]:
                self._grid[cell[0]][cell[1]][1] = False
                self._grid[neighbor[0]][neighbor[1]][0] = False
            
            # +---+
            # | N |
            # +---+
            # | C |
            # +---+
            else:
                self._grid[cell[0]][cell[1]][0] = False
                self._grid[neighbor[0]][neighbor[1]][1] = False
                
        elif cell[0] == neighbor[0]:  # If both cells share the same row.
            # +---+---+
            # | C | N |
            # +---+---+
            if cell[1] < neighbor[1]:
                self._grid[cell[0]][cell[1]][2] = False
                self._grid[neighbor[0]][neighbor[1]][3] = False

            # +---+---+
            # | N | C |
            # +---+---+
            else:
                self._grid[cell[0]][cell[1]][3] = False
                self._grid[neighbor[0]][neighbor[1]][2] = False



import numpy as np


class MazeGenerator:
    """A base class for maze-generating algorithms."""

    def __init__(self):
        self._grid = None

    def _remove_wall(self, cell: tuple[int, int], neighbor: tuple[int, int]):
        """Remove the wall between a cell and its neighbor.

        Parameters
        ----------
        cell : tuple[int, int]
            The location of the cell.
        neighbor : tuple[int, int]
            The location of the cell's neighbor.
        """
        if cell[1] == neighbor[1]:  # If both cells share the same column.
            # +---+
            # | C |
            # +---+
            # | N |
            # +---+
            if cell[0] < neighbor[0]:
                self._grid[cell[0]][cell[1]][1] = False
                self._grid[neighbor[0]][neighbor[1]][0] = False
            
            # +---+
            # | N |
            # +---+
            # | C |
            # +---+
            else:
                self._grid[cell[0]][cell[1]][0] = False
                self._grid[neighbor[0]][neighbor[1]][1] = False

        elif cell[0] == neighbor[0]:  # If both cells share the same row.
            # +---+---+
            # | C | N |
            # +---+---+
            if cell[1] < neighbor[1]:
                self._grid[cell[0]][cell[1]][2] = False
                self._grid[neighbor[0]][neighbor[1]][3] = False

            # +---+---+
            # | N | C |
            # +---+---+
            else:
                self._grid[cell[0]][cell[1]][3] = False
                self._grid[neighbor[0]][neighbor[1]][2] = False

    def _initiate_grid(self, rows: int, columns: int, walls: bool = False):
        """Initiate a two-dimensional list of each cell's wall data.

        The wall data is a list consisting of four Boolean values in NSEW order.

        Parameters
        ----------
        rows : int
            The total number of rows of the maze.
        columns : int
            The total number of columns of the maze.
        walls : bool
            Whether to initiate all the cell with walls. Defaults to ``False``.
        """

        self._grid = np.full((rows, columns, 4), [walls] * 4)

    def generate(self, rows: int, columns: int):
        """An abstract method to generate a maze.

        Parameters
        ----------
        rows : int
            The total number of rows of the maze.
        columns : int
            The total number of columns of the maze.

        Raises
        ------
        NotImplementedError
            If the method has not been implemented in a subclass.
        """
        raise NotImplementedError(
            "The 'generate()' method must be implemented in the subclass."
        )



import random
import sys

import numpy as np

from .maze_generator import MazeGenerator


class RecursiveBacktracking(MazeGenerator):
    """A maze-generating algorithm that creates a perfect maze using a randomized version of depth-first search."""

    def __init__(self):
        super().__init__()
        self._visited = None

    def _visit_cell(self, row, column):
        """Visit a cell using recursive backtracking.

        Parameters
        ----------
        row : int
            The row of the cell to be visited.
        column : int
            The column of the cell to be visited.
        """
        self._visited.add((row, column))  # Mark current cell as visited

        # Iterate over the four directions in a randomized order
        for row_delta, column_delta in random.sample(
            ((-1, 0), (0, 1), (1, 0), (0, -1)), k=4
        ):
            # Get a random neighbor's row and column
            neighbor = (row + row_delta, column + column_delta)

            # See if the neighbor exists
            if (
                neighbor[0] < 0
                or neighbor[0] >= len(self._grid)
                or neighbor[1] < 0
                or neighbor[1] >= len(self._grid[0])
            ):
                continue

            # See if the neighbor is already visited
            if neighbor in self._visited:
                continue

            # Remove the wall between the current cell and the selected neighbor
            self._remove_wall((row, column), neighbor)

            # Visit the neighbor recursively
            self._visit_cell(neighbor[0], neighbor[1])

    def generate(self, rows: int, columns: int) -> np.ndarray:
        """Generate a maze.

        Parameters
        ----------
        rows : int
            The total number of rows of the maze.
        columns : int
            The total number of columns of the maze.

        Returns
        -------
        numpy.ndarray
            A two-dimensional array of cells representing a rectangular maze.
        """
        sys.setrecursionlimit(max(1000, rows * columns))
        self._visited = set()
        self._initiate_grid(rows, columns, walls=True)
        self._visit_cell(random.randrange(rows), random.randrange(columns))
        return self._grid
